#include "die.h"
#include "ev.h"
#include <stdlib.h>
#include <string.h>
#include "analyze_output.h"
#include "mtanalyze.h"
#include "analyze_mutex_deadlocks.h"
#include "preanalyze.h"

class LockHistory;

class LockStack {
	struct Lock {
		Object *mtx;
		const EventRecord *ev;
	};
	Lock embedded_lock[5];
	Lock *lock;
	size_t locks;
public:
	LockStack() : lock(embedded_lock), locks(0) {}
	~LockStack() { if(lock!=embedded_lock) delete[] lock; }
	LockStack(const LockStack &ls);
	LockStack &operator=(const LockStack &ls);
	
	void push(Object *mtx, const EventRecord *ev, LockHistory *lh);
	void pop(Object *mtx);
	size_t size() const { return locks; }
	Object *mtx(size_t i) { return lock[i].mtx; }
	const EventRecord *ev(size_t i) { return lock[i].ev; }
	
	bool isPrefixOf(const LockStack &ls) const;
	bool operator==(const LockStack &ls) const;
	
	void compress();
};

class LockHistory {
	LockStack *e;
	size_t es;
	size_t e_alloc;
public:
	LockHistory() : e(0), es(0), e_alloc(0) {}
	~LockHistory() { delete[] e; }
	void add(const LockStack &ls);
	size_t size() const { return es; }
	LockStack& operator[](size_t i) { return e[i]; }
	void remove(size_t i);
	
	void compress();
};



LockStack::LockStack(const LockStack &ls)
  : lock(embedded_lock), locks(0)
{
	*this = ls;
}

LockStack& LockStack::operator=(const LockStack &ls) {
	if(&ls==this) return *this;
	if(lock!=embedded_lock) delete[] lock;
	locks = ls.locks;
	if(locks<=5) {
		lock = embedded_lock;
	} else {
		lock = new Lock[locks];
	}
	for(size_t i=0; i<ls.locks; i++)
		lock[i] = ls.lock[i];
	return *this;
}
	
void LockStack::push(Object *mtx, const EventRecord *ev, LockHistory *lh) {
	if(lock!=embedded_lock || locks>=5) {
		Lock *nl = new Lock[locks+1];
		for(size_t i=0; i<locks; i++)
			nl[i] = lock[i];
		if(lock!=embedded_lock) delete[] lock;
		lock = nl;
	}
	lock[locks].mtx = mtx;
	lock[locks].ev = ev;
	locks++;
	
	if(locks>1)
		lh->add(*this);
}

void LockStack::pop(Object *mtx) {
	if(lock[locks-1].mtx==mtx) {
		//avg. case
		locks--;
		return;
	} else {
		for(size_t i = locks; i--!=0; ) {
			if(lock[i].mtx == mtx) {
				memmove(lock+i, lock+i+1, sizeof(Lock)*(locks-i-1));
				locks--;
				return;
			}
		}
	}
}
	
bool LockStack::isPrefixOf(const LockStack &ls) const {
	if(ls.locks<locks) return false;
	for(size_t i=0; i<locks; i++) {
		if(lock[i].mtx!=ls.lock[i].mtx)
			return false;
	}
	return true;
}

bool LockStack::operator==(const LockStack &ls) const {
	if(locks!=ls.locks) return false;
	return isPrefixOf(ls);
}

void LockStack::compress() {
	//remove duplicated mutexes
	for(size_t i=0; i<locks; i++) {
		for(size_t j=i+1; j<locks; ) {
			if(lock[i].mtx==lock[j].mtx) {
				for(size_t k=j; k<locks-1; k++)
					lock[k]=lock[k+1];
				locks--;
			} else
				j++;
		}
	}
}


void LockHistory::add(const LockStack &ls) {
	if(es==e_alloc) {
		size_t nea = e_alloc + 32;
		LockStack *ne = new LockStack[nea];
		for(size_t i=0; i<es; i++)
			ne[i] = e[i];
		delete[] e;
		e = ne;
		e_alloc = nea;
	}
	
	size_t back = es<5?0:es-5;
	for(size_t i=es; i--!=back; ) {
		if(ls.isPrefixOf(e[i]))
			return;
		if(e[i].isPrefixOf(ls)) {
			e[i] = ls;
			return;
		}
	}
	
	e[es++] = ls;
}

void LockHistory::remove(size_t i) {
	for( ; i<es-1; i++)
		e[i] = e[i+1];
	es--;
}

void LockHistory::compress() {
	for(size_t i=0; i<es; i++)
		e[i].compress();
	
	for(size_t i=es; i-->0; ) {
		bool drop=false;
		for(size_t j=0; j<es; j++) {
			if(j!=j && e[i].isPrefixOf(e[j])) {
				drop=true;
				break;
			}
		}
		if(drop) {
			for(size_t j=i; j<es-1; j++)
				e[j] = e[j+1];
			es--;
		}
	}
}

class ThreadLockStackMap {
public:
	struct Element {
		Object *thr;
		LockHistory lh;
		LockStack ls;
		Element *next;
	};
	Element *first;

	ThreadLockStackMap() : first(0) {}
	~ThreadLockStackMap();
	LockStack& stackOf(Object *thr) { return find(thr)->ls; }
	LockHistory& historyOf(Object *thr) { return find(thr)->lh; }

	Element *find(Object *thr);
	
	void compress();
	bool empty() const { return first==0; }
	void counts(unsigned *threads, unsigned *stacks, unsigned *locks);
};

ThreadLockStackMap::~ThreadLockStackMap() {
	Element *p=first;
	while(p) {
		Element *n=p->next;
		delete p;
		p=n;
	}
}

ThreadLockStackMap::Element *ThreadLockStackMap::find(Object *thr) {
	Element *p=first;
	Element *prev=0;
	while(p && p->thr!=thr) {
		prev=p;
		p=p->next;
	}
	if(p && p==first) return p;
	if(!p) {
		p = new Element;
		p->thr = thr;
		p->next=first;
		first=p;
		return p;
	}
	prev->next = p->next;
	p->next = first;
	first = p;
	return p;
}

void ThreadLockStackMap::compress() {
	for(Element *p=first; p; p=p->next)
		p->lh.compress();
}

void ThreadLockStackMap::counts(unsigned *threads, unsigned *stacks, unsigned *locks) {
	size_t t=0,s=0,l=0;
	for(Element *p=first; p; p=p->next) {
		t++;
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); i++) {
			s++;
			l += lh[i].size();
		}
	}
	*threads = (unsigned)t;
	*stacks = (unsigned)s;
	*locks = (unsigned)l;
}




static void build_threadlockstackmap(const EventRecord ev[], size_t events,
                                     EventObjectPointers *eop,
		        	     ThreadLockStackMap &map)
{
	//scan through events, building lock stacks for threads
	for(size_t i=0; i<events; i++) {
		const EventRecord *ep=ev+i;
		if(!eop[i].mtx) continue;
		
		LockStack &ls = map.stackOf(eop[i].thr);
		LockHistory &lh = map.historyOf(eop[i].thr);
		
		switch(ep->event) {
			case ev_mutex_lock:
			case ev_cond_waited:
			case ev_cond_timedwait_failed:
				ls.push(eop[i].mtx,ep,&lh);
				break;
			case ev_mutex_trylock:
				//nop - postponed to ...locked
				break;
			case ev_mutex_locked:
				if(ls.ev(ls.size()-1)->event==ev_mutex_trylock)
					ls.push(eop[i].mtx,ep,&lh);
				break;
			case ev_mutex_unlock:
			case ev_cond_wait:
			case ev_cond_timedwait:
				ls.pop(eop[i].mtx);
				break;
		}
	}
}


class MutexSet {
	struct E {
		E *next;
		Object *mtx;
	};
	enum { N=37 };
	E *e[N];
	static unsigned hash(Object *mtx);
public:
	MutexSet();
	~MutexSet();
	void add(Object *mtx);
	bool exists(Object *mtx) const;
	void remove(Object *mtx);
};

inline unsigned MutexSet::hash(Object *mtx) {
	return (((unsigned)(unsigned long)mtx)>>2)%N;
}

MutexSet::MutexSet() {
	memset(e,0,sizeof(e));
}

MutexSet::~MutexSet() {
	for(int i=0; i<N; i++) {
		E *p=e[i];
		while(p) {
			E *n=p->next;
			delete p;
			p=n;
		}
	}
}

void MutexSet::add(Object *mtx) {
	unsigned hi=hash(mtx);
	E *p=e[hi];
	while(p) {
		if(p->mtx==mtx) return;
		p=p->next;
	}
	p = new E;
	p->mtx = mtx;
	p->next = e[hi];
	e[hi] = p;
}

bool MutexSet::exists(Object *mtx) const {
	unsigned hi=hash(mtx);
	E *p=e[hi];
	while(p) {
		if(p->mtx==mtx) return true;
		p=p->next;
	}
	return false;
}

void MutexSet::remove(Object *mtx) {
	unsigned hi=hash(mtx);
	E *p=e[hi];
	E *pp=0;
	while(p) {
		if(p->mtx==mtx) {
			if(pp)
				pp->next = p->next;
			else
				e[hi] = p->next;
			delete p;
			return;
		}
		pp=p;
		p=p->next;
	}
}




static bool eliminate_inner_locks(ThreadLockStackMap &map) {
	MutexSet inner_mutexes;
	//find mutexes on innermost levels
	for(ThreadLockStackMap::Element *p=map.first; p; p=p->next) {
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); i++) {
			LockStack &ls = lh[i];
			if(ls.size()>0)
				inner_mutexes.add(ls.mtx(ls.size()-1));
		}
	}
	//remove mutexes used in not-innermost levels
	for(ThreadLockStackMap::Element *p=map.first; p; p=p->next) {
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); i++) {
			LockStack &ls = lh[i];
			for(size_t j=0; j<ls.size()-1; j++)
				inner_mutexes.remove(ls.mtx(j));
		}
	}
	//what we have left are mutexes used only at innermost level
	bool any_removed=false;
	for(ThreadLockStackMap::Element *p=map.first; p; p=p->next) {
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); i++) {
			LockStack &ls = lh[i];
			if(ls.size()>0) {
				Object *mtx = ls.mtx(ls.size()-1);
				if(inner_mutexes.exists(mtx)) {
					ls.pop(mtx);
					any_removed=true;
				}
			}
		}
	}
	return any_removed;
}


static bool eliminate_outer_locks(ThreadLockStackMap &map) {
	MutexSet outer_mutexes;
	//find mutexes on outermost levels
	for(ThreadLockStackMap::Element *p=map.first; p; p=p->next) {
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); i++) {
			LockStack &ls = lh[i];
			if(ls.size()>0)
				outer_mutexes.add(ls.mtx(0));
		}
	}
	//remove mutexes used in not-outermost levels
	for(ThreadLockStackMap::Element *p=map.first; p; p=p->next) {
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); i++) {
			LockStack &ls = lh[i];
			for(size_t j=1; j<ls.size(); j++)
				outer_mutexes.remove(ls.mtx(j));
		}
	}
	//what we have left are mutexes used only at outermost level
	bool any_removed=false;
	for(ThreadLockStackMap::Element *p=map.first; p; p=p->next) {
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); i++) {
			LockStack &ls = lh[i];
			if(ls.size()>0) {
				Object *mtx = ls.mtx(ls.size()-1);
				if(outer_mutexes.exists(mtx)) {
					ls.pop(mtx);
					any_removed=true;
				}
			}
		}
	}
	return any_removed;
}

static void eliminate_1level_locks(ThreadLockStackMap &map) {
	for(ThreadLockStackMap::Element *p=map.first; p; p=p->next) {
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); ) {
			LockStack &ls = lh[i];
			if(ls.size()<=1)
				lh.remove(i);
			else
				i++;
		}
	}
}

static void eliminate_threads(ThreadLockStackMap &map) {
	ThreadLockStackMap::Element *p=map.first;
	ThreadLockStackMap::Element *prev = 0;
	while(p) {
		if(p->lh.size()==0) {
			if(prev)
				prev->next = p->next;
			else
				map.first = p->next;
			ThreadLockStackMap::Element *n=p->next;
			delete p;
			p=n;
		} else {
			prev = p;
			p=p->next;
		}
	}
	
}


static void print_deadlocks(ThreadLockStackMap &map) {
	//todo: use dyn.equiv.rel. to sort into deadlock groups
	fprintf(fp_output,"(This program cannot present the deadlocks in a easily comprehensible format yet)\n");
	
	for(ThreadLockStackMap::Element *p=map.first; p; p=p->next) {
		LockHistory &lh = p->lh;
		for(size_t i=0; i<lh.size(); i++) {
			LockStack &ls = lh[i];
			OutputParagraphStart();
			
			fprintf(fp_output,"Thread %ld ", p->thr->addr);
			if(p->thr->name_event)
				fprintf(fp_output,"(%s) ",p->thr->name_event->name);
			if(p->thr->create_event || p->thr->destroy_event) {
				fprintf(fp_output," Lifetime: ");
				if(p->thr->create_event)
					fprintf(fp_output,"%.6f ",p->thr->create_event->timestamp);
				else
					fprintf(fp_output,"program start ");
				fprintf(fp_output,"to ");
				if(p->thr->destroy_event)
					fprintf(fp_output,"%.6f",p->thr->destroy_event->timestamp);
				else
					fprintf(fp_output,"program end");
								if(p->thr->create_event)
				fprintf(fp_output,"\n");
			} else
				fprintf(fp_output," Lifetime: entire program\n");
			
			fprintf(fp_output,"Time of deadlock possibility: %.6f\n", ls.ev(ls.size()-1)->timestamp);
			
			fprintf(fp_output,"Lock stack:\n");
			for(size_t j=0; j<ls.size(); j++) {
				Object *m = ls.mtx(j);
				fprintf(fp_output,"%08lx ", m->addr);
				if(m->name_event)
					fprintf(fp_output,"(%s) ",m->name_event->name);
				if(m->create_event || m->create_event) {
					if(m->create_event) fprintf(fp_output,"%.6f to ",m->create_event->timestamp);
					else fprintf(fp_output,"static to ");
					if(m->destroy_event) fprintf(fp_output,"%.6f",m->destroy_event->timestamp);
					else fprintf(fp_output,"program end");
				} else
					fprintf(fp_output,"(static mutex) ");
				fprintf(fp_output,"\n");
			}
			
			OutputParagraphEnd();
		}
	}
}

void analyze_mutex_deadlocks(const EventRecord ev[], size_t events,
                             Object *, size_t ,
                             EventObjectPointers *eop)
{
	ThreadLockStackMap map;
	unsigned threads,stacks,locks;
	fprintf(fp_output,"Building lock traces\n");
	build_threadlockstackmap(ev,events,eop, map);

	map.counts(&threads,&stacks,&locks);
	fprintf(fp_output,"To be examined: %u threads, %u traces and %u lock operations\n",threads,stacks,locks);
	
	//Compress the traces to produce more managable data
	if(stacks==0) {
		fprintf(fp_output,"Skipping analysis since deadlocks are not possible\n");
		return;
	}
	
	fprintf(fp_output,"Eliminating redundant traces\n");
	map.compress();

	map.counts(&threads,&stacks,&locks);
	fprintf(fp_output,"To be examined: %u threads, %u traces and %u lock operations\n",threads,stacks,locks);
	if(stacks==0) {
		fprintf(fp_output,"Skipping analysis since deadlocks are not possible\n");
		return;
	}
	
	//
	fprintf(fp_output,"Eliminating non-essential locks:\n");
	int iteration=1;
	for(;;) {
		fprintf(fp_output,"Iteration %d\n",iteration);
		map.counts(&threads,&stacks,&locks);
		fprintf(fp_output,"To be examined: %u threads, %u traces and %u lock operations\n",threads,stacks,locks);
		fprintf(fp_output,"Eliminating locks used only at inner level\n");
		bool b1=eliminate_inner_locks(map);
		fprintf(fp_output,"Eliminating locks used only at outer level\n");
		bool b2=eliminate_outer_locks(map);
		fprintf(fp_output,"Eliminating 1-level lock traces\n");
		eliminate_1level_locks(map);
		fprintf(fp_output,"Eliminating threads with no interesting lock patterns\n");
		eliminate_threads(map);
		fprintf(fp_output,"Reducing traces\n");
		map.compress();
		if(!b1 && !b2) break;
		iteration++;
	}
	map.counts(&threads,&stacks,&locks);
	fprintf(fp_output,"Still left : %u threads, %u traces and %u lock operations\n",threads,stacks,locks);
	if(threads==0) {
		fprintf(fp_output,"No deadlocks found\n");
		return;
	}
	
	fprintf(fp_output,"Deadlocks found (real and/or potential)\n");
	print_deadlocks(map);
}

