#include "mtanalyze.h"
#include "logread.h"
#include "preanalyze.h"
#include "analyze_mutex.h"
#include "analyze_condition.h"
#include "analyze_thread.h"
#include "stat_thread.h"
#include "stat_mutex.h"
#include "stat_condition.h"
#include "msgconfig.h"
#include "analyze_output.h"
#include "version.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

bool allow_recursive_mutexes=false;
#ifdef __LINUX__
//This setting depends on the version of LinuxThreads. There is no #define to
//test on for this, if we assume that glibc 2.2+ means a linuxthreads 0.9+ ...
//(a bad assumption but it is all we have got)
#if (__GLIBC__ + 0 > 2) || ( (__GLIBC__ + 0 == 2 ) && (__GLIBC_MINOR__ + 0 >= 3))
bool implicit_cond_mutex_ops=true;
#else
bool implicit_cond_mutex_ops=false;
#endif
#else
bool implicit_cond_mutex_ops=true;
#endif
double warn_mutex_contention_ratio = 0.1;
double warn_mutex_long_lock_time = 1.0;
double warn_mutex_long_hold_time = 1.0;
double warn_mutex_longer_lock_time = 10.0;
double warn_failed_trylock_ratio = 0.2;
unsigned long warn_deep_locks=5;
unsigned long analyze_singlethread_mutex_threshold=0;
unsigned warn_mutex_many_waiters_threshold=3;
bool perform_deadlock_analysis=true;

size_t dump_hot_mutex_count=5;
bool dump_mutexes=false;

bool dump_conditions=false;
size_t dump_hot_condition_count=5;
unsigned long warn_signal_without_waiters_min_ops=100;
double warn_signal_without_waiters_ratio=0.3;
unsigned warn_thundering_herd_max_intervening_ops=2;
double warn_thundering_herd_max_intervening_time=0.01;
double warn_thundering_herd_ratio=0.6;

unsigned warn_concurrent_threads_threshold=200;
double warn_shortlived_thread_ratio=0.1;
double warn_shortlived_threads_ratio=0.5;
bool dump_threads=false;

bool generate_statistics_mutex=true;
bool generate_statistics_condition=true;
bool generate_statistics_thread=true;

static const char *input_filename="mta.log";
static char *output_filename=0;


static void print_help() {
printf("mtanalyze v0.1 - analyze thread, mutexes and condition use\n"
       "\n"
       "Usage:\n"
       "mtanalyze [option...]\n"
      );
printf("\nOptions:\n");

printf("  -of filename Set output filename. Default is stdout\n");
printf("  -if filename Set event log filename. Default is mta.log\n");
printf("\n");
printf("  +r    -r     Enable/disable recursive mutexes\n"
       "               If recursive mutexes are not used more strict analysis can\n"
       "               be done and more errors can be caught.\n"
       "               (default: %s)\n", allow_recursive_mutexes?"enabled":"disabled");
printf("\n");
printf("  +i    -i     Enable/disable implicit pthread_cond_wait mutex operations\n"
       "               Some pthread libraries calls pthread_mutex_lock/unlock\n"
       "               internally when pthread_cond_wait() is called.\n"
       "               (default: %s)\n", implicit_cond_mutex_ops?"enabled":"disabled");
printf("\n");
printf("  +w x  -w x   Enable/disable warning x\n");
printf("\n");
printf("  -wmcr x      Warn mutex contention ratio\n"
       "               If the ratio of contended lock operations is above this level\n"
       "               a warning is generated.\n"
       "               (default: %.2f)\n", warn_mutex_contention_ratio);
printf("  -wmllt x     Warn mutex long lock time\n"
       "               If a lock operation takes more than the specified time a warning\n"
       "               is generated.\n"
       "               (default: %.2f seconds)\n", warn_mutex_long_hold_time);
printf("  -wmlht x     Warn mutex long hold time\n"
       "               Generate warning if a mutex is held for more than the\n"
       "               specified time.\n"
       "               (default: %.2f)\n", warn_mutex_long_hold_time);
printf("  -wmLlt x     Warn mutex longer-than-usual factor\n"
       "               Generate warning if a mutex is held more than <x> times the\n"
       "               average hold time.\n"
       "               (default: %.1f)\n", warn_mutex_longer_lock_time);
printf("  -wmft x      Threshold for warning about failed pthread_mutex_trylock\n"
       "               If the ratio of failed pthread_mutex_trylock() operations\n"
       "               exceeds this value a warning is generated.\n"
       "               (default: %.2f)\n", warn_failed_trylock_ratio);
printf("  -wmdl x      Threshold for warning about deep locks\n"
       "               If a thread has more than <x> locks a warning is generated.\n"
       "               (default: %lu)\n", warn_deep_locks);
printf("  -wmsm x      Threshold for analyzing single-thread mutexes\n"
       "               If a mutex has more than <x> oprations performed upon it it is\n"
       "               subject to single-thread-use analysis.\n"
       "               (default: %lu)\n", analyze_singlethread_mutex_threshold);
printf("  -wmmw x      Threshold for warning about many waiters on a mutex\n"
       "               If the number of threads waiting for a mutex exceeds this value\n"
       "               a warning is generated.\n"
       "               (default: %u)\n", warn_mutex_many_waiters_threshold);
printf("\n");
printf("  -wcswwmo x   Threshold for warning about pthread_cond_signal without waiting\n"
       "               threads. Minimum pthread_cond_signal() calls on a condition\n"
       "               without any waiting threads before generating a warning.\n"
       "               (default: %lu)\n", warn_signal_without_waiters_min_ops);
printf("  -wswwr   x   Threshold ratio before warning about pthread_cond_signal without\n"
       "               waiting threads. If the number of pthread_cond_signal() calls\n"
       "               without any waiting threads exceeds this ratio a warning is\n"
       "               generated.\n"
       "               (default: %.2f)\n", warn_signal_without_waiters_ratio);
printf("  -wthmio x    Warn about \"thundering herd\" - maximum intervening operations\n"
       "               If a thread performs more than the specified number of\n"
       "               pthread calls after being awakened it is assumed to perform real\n"
       "               work.\n"
       "               (default: %u)\n", warn_thundering_herd_max_intervening_ops);
printf("  -wthmit x    Warn about \"thundering herd\" - maximum intervening time\n"
       "               If a thread after being awakened spends more than the\n"
       "               specified time before waiting again it is assumed to perform\n"
       "               real work\n"
       "               (default: %.3f seconds)\n", warn_thundering_herd_max_intervening_time);
printf("  -wthmr x     Warn about \"thundering herd\" - minimum ratio\n"
       "               If more than the specified ratio of awakened threads seem to\n"
       "               not perform any real work a warning is genereated\n"
       "               (default: %.2f)\n", warn_thundering_herd_ratio);
printf("\n");
printf("  -hm x        Dump <x> most used mutexes\n"
       "               (default: %lu)\n", (unsigned long)dump_hot_mutex_count);
printf("  -hc x        Dump <x> most used conditions\n"
       "               (default: %lu)\n", (unsigned long)dump_hot_condition_count);
printf("\n");
printf("  +ad  -ad     Enable/disable deadlock analysis\n"
       "               (default: %s)\n", perform_deadlock_analysis?"enabled":"disabled");
printf("\n");  
printf("  +dm  -dm     Dump all mutexes\n"
       "               (default: %s)\n", dump_mutexes?"enabled":"disabled");
printf("  +dc  -dc     Dump all conditions\n"
       "               (default: %s)\n", dump_conditions?"enabled":"disabled");
printf("  +dt  -dt     Dump all threads\n"
       "               (default: %s)\n", dump_threads?"enabled":"disabled");
printf("\n");
printf("  +sm  -sm     Generate mutex statistics\n"
       "               (default: %s)\n", generate_statistics_mutex?"enabled":"disabled");
printf("  +sc  -sc     Generate condition statistics\n"
       "               (default: %s)\n", generate_statistics_condition?"enabled":"disabled");
printf("  +st  -st     Generate thread statistics\n"
       "               (default: %s)\n", generate_statistics_thread?"enabled":"disabled");
printf("\n");
printf("Return value\n"
       "  The return value is zero if no fatal errors were found. It is non-zero if a\n"
       "  fatal error were found\n");
}

static void parseOptions(int argc, char * const argv[]) {
	for(int i=1; i<argc; i++) {
		const char *a=argv[i];
		if(strcmp(a,"--version")==0) {
			printf("mta version %s\n",VERSION);
			exit(0);
		} else if(strcmp(a,"-if")==0) {
			if(i<argc-1) {
				i++;
				input_filename = argv[i];
			}
		} else if(strcmp(a,"-of")==0) {
			if(i<argc-1) {
				i++;
				output_filename=argv[i];
			}
		} else if(strcmp(a,"+r")==0) {
			allow_recursive_mutexes=true;
		} else if(strcmp(a,"-r")==0) {
			allow_recursive_mutexes=false;
		} else if(strcmp(a,"+i")==0) {
			implicit_cond_mutex_ops=true;
		} else if(strcmp(a,"-i")==0) {
			implicit_cond_mutex_ops=false;
		} else if(strcmp(a,"+w")==0 || strcmp(a,"-w")==0) {
			if(i<argc-1) {
				bool enable = strcmp(a,"+w")==0;
				i++;
		        	int num = (int)strtol(argv[i],0,0);
				enableMessage(num,enable);
			}
		} else if(strcmp(a,"-wmcr")==0) {
			if(i<argc-1) {
				i++;
				warn_mutex_contention_ratio = strtod(argv[i],0);
			}
		} else if(strcmp(a,"-wmllt")==0) {
			if(i<argc-1) {
				i++;
				warn_mutex_long_lock_time = strtod(argv[i],0);
			}
		} else if(strcmp(a,"-wmlht")==0) {
			if(i<argc-1) {
				i++;
				warn_mutex_long_hold_time = strtod(argv[i],0);
			}
		} else if(strcmp(a,"-wmLlt")==0) {
			if(i<argc-1) {
				i++;
				warn_mutex_longer_lock_time = strtod(argv[i],0);
			}
		} else if(strcmp(a,"-wmft")==0) {
			if(i<argc-1) {
				i++;
				warn_failed_trylock_ratio = strtod(argv[i],0);
			}
		} else if(strcmp(a,"-wmdl")==0) {
			if(i<argc-1) {
				i++;
				warn_deep_locks = strtol(argv[i],0,0);
			}
		} else if(strcmp(a,"-wmsm")==0) {
			if(i<argc-1) {
				i++;
				analyze_singlethread_mutex_threshold = strtol(argv[i],0,0);
			}
		} else if(strcmp(a,"-wmmw")==0) {
			if(i<argc-1) {
				i++;
				warn_mutex_many_waiters_threshold = (unsigned)strtol(argv[i],0,0);
			}
		} else if(strcmp(a,"-wcswwmo")==0) {
			if(i<argc-1) {
				i++;
				warn_signal_without_waiters_min_ops = strtol(argv[i],0,0);
			}
		} else if(strcmp(a,"-wswwr")==0) {
			if(i<argc-1) {
				i++;
				warn_signal_without_waiters_ratio = strtod(argv[i],0);
			}
		} else if(strcmp(a,"-wthmio")==0) {
			if(i<argc-1) {
				i++;
				warn_thundering_herd_max_intervening_ops = (unsigned)strtol(argv[i],0,0);
			}
		} else if(strcmp(a,"-wthmit")==0) {
			if(i<argc-1) {
				i++;
				warn_thundering_herd_max_intervening_time = strtod(argv[i],0);
			}
		} else if(strcmp(a,"-wthmr")==0) {
			if(i<argc-1) {
				i++;
				warn_thundering_herd_ratio = strtod(argv[i],0);
			}
		} else if(strcmp(a,"-hm")==0) {
			if(i<argc-1) {
				i++;
				dump_hot_mutex_count = strtol(argv[i],0,0);
			}
		} else if(strcmp(a,"-hc")==0) {
			if(i<argc-1) {
				i++;
				dump_hot_condition_count = strtol(argv[i],0,0);
			}
		} else if(strcmp(a,"+ad")==0) {
			perform_deadlock_analysis=true;
		} else if(strcmp(a,"-ad")==0) {
			perform_deadlock_analysis=false;
		} else if(strcmp(a,"+dm")==0) {
			dump_mutexes=true;
		} else if(strcmp(a,"-dm")==0) {
			dump_mutexes=false;
		} else if(strcmp(a,"+dc")==0) {
			dump_conditions=true;
		} else if(strcmp(a,"-dc")==0) {
			dump_conditions=false;
		} else if(strcmp(a,"+dt")==0) {
			dump_threads=true;
		} else if(strcmp(a,"-dt")==0) {
			dump_threads=false;
		} else if(strcmp(a,"+sm")==0) {
			generate_statistics_mutex=true;
		} else if(strcmp(a,"-sm")==0) {
			generate_statistics_mutex=false;
		} else if(strcmp(a,"+sc")==0) {
			generate_statistics_condition=true;
		} else if(strcmp(a,"-sc")==0) {
			generate_statistics_condition=false;
		} else if(strcmp(a,"+st")==0) {
			generate_statistics_thread=true;
		} else if(strcmp(a,"-st")==0) {
			generate_statistics_thread=false;
		}
	}
}

int main(int argc, char *argv[]) {
	if(argc==2 && (strcmp(argv[1],"--help")==0 || strcmp(argv[1],"--usage")==0)) {
		print_help();
		return 0;
	}
	parseOptions(argc,argv);
	EventRecord *ev=0;
	size_t evs=0;
	printf("Reading event log..."); fflush(stdout);
	if(!ReadLog(input_filename,&ev,&evs)) {
		fprintf(stderr,"Could not read log \n");
		return -1;
	}
	printf("\nEvent log read (%ld events)\n",(long)evs);
	
	if(output_filename) {
		fp_output = fopen(output_filename,"w");
		if(!fp_output) {
			perror(output_filename);
			return EXIT_FAILURE;
		}
	}
	
	Object *mtx=0;
	size_t mtxs=0;
	Object *cond=0;
	size_t conds=0;
	Object *thr=0;
	size_t thrs=0;
	EventObjectPointers *eop=0;
	preAnalyze(ev,evs, &mtx,&mtxs, &cond,&conds, &thr,&thrs, &eop);
	
	analyze_mutex_use(ev,evs, mtx,mtxs, eop);
	
	analyze_condition_use(ev,evs, cond,conds, eop);
	
	analyze_thread_use(ev,evs, thr,thrs, eop);
	
	if(generate_statistics_mutex)
		stat_mutex_use(ev,evs, mtx,mtxs, eop);
	if(generate_statistics_condition)
		stat_condition_use(ev,evs, cond,conds, eop);
	if(generate_statistics_thread)
		stat_thread_use(ev,evs, thr,thrs, eop);
	
	return 0;
}
