#define __STDC_FORMAT_MACROS
#include "preanalyze.h"
#include "aohash.h"
#include "ahash.h"
#include "ev.h"
#include <stdlib.h>
#include <string.h>
#include "die.h"
#include "analyze_output.h"

//Class for handling expanding arrays and re-fixing pointers
class ObjectMemoryHandler {
	Object **obj;
	size_t *objs;
	EventObjectPointers *eop;
	const size_t events;
	AddressObjectHashTable &h;
	
	size_t objs_allocated;
	static void fixPointer(Object *&p, Object *old_base, Object *new_base) {
		p = new_base + (p-old_base);
	}
	void expand();
public:
	ObjectMemoryHandler(Object **obj_, size_t *objs_, EventObjectPointers eop_[], size_t events_, AddressObjectHashTable &h_)
	  : obj(obj_),
	    objs(objs_),
	    eop(eop_),
	    events(events_),
	    h(h_),
	    objs_allocated(0)
	  {}
	
	Object *addObject();
};


void ObjectMemoryHandler::expand()
{
	//expand array, and fix pointers
	Object *old_obj_begin = *obj;
	size_t new_alloc_size = objs_allocated + 256;
	*obj = (Object*)realloc(*obj, sizeof(Object)*(new_alloc_size));
	if(old_obj_begin && *obj!=old_obj_begin) {
		//fix pointers in eop[]
		Object *old_obj_end = old_obj_begin + *objs;
		for(size_t e=0; e<events; e++) {
			if(eop[e].mtx>=old_obj_begin  && eop[e].mtx<old_obj_end)  fixPointer(eop[e].mtx,old_obj_begin,*obj);
			if(eop[e].cond>=old_obj_begin && eop[e].cond<old_obj_end) fixPointer(eop[e].cond,old_obj_begin,*obj);
			if(eop[e].thr>=old_obj_begin  && eop[e].thr<old_obj_end)  fixPointer(eop[e].thr,old_obj_begin,*obj);
		}
		h.refixPointers(old_obj_begin,old_obj_end,*obj);
	}
	objs_allocated = new_alloc_size;
}

Object *ObjectMemoryHandler::addObject()
{
	if(*objs>=objs_allocated)
		expand();
	Object *o = (*obj)+*objs;
	++*objs;
	o->create_event = 0;
	o->destroy_event = 0;
	o->name_event = 0;
	return o;
}


static void findMutexes(const EventRecord ev[], size_t events,
                        Object **o_mtx, size_t *mtxs,
                        EventObjectPointers eop[]
                       )
{
	AddressObjectHashTable h;
	SeenAddressTable sat;
	ObjectMemoryHandler omh(o_mtx,mtxs,eop,events,h);
	for(size_t i=0; i<events; i++) {
		const EventRecord *ep=ev+i;
		if(ep->mtx_addr==0) continue;
		Object *obj = h.find(ep->mtx_addr);
		bool used_address = sat.exists(ep->mtx_addr);
		if(!used_address) sat.add(ep->mtx_addr);
		
		switch(ep->event) {
		case ev_mutex_name: {
			if(obj) {
				if(obj->name_event)
					warn(ep,MUTEXALREADYNAMED,
					     "Mutex has already been named '%s'\n",
					     obj->name_event->name
					    );
				else
				 	obj->name_event = ep;
			} else {
				obj = omh.addObject();
				obj->addr = ep->mtx_addr;
				obj->name_event = ep;
				h.add(ep->mtx_addr,obj);
			}
			break;
		}
		case ev_mutex_create: {
			//object known ?
			//  => bad
			if(obj)
				die(ep,MUTEXINITEADDRESS,"pthread_mutex_init() called for address where a mutex already exists\n");
			break;
		}
		case ev_mutex_created: {
			//object known?
			//  => bad
			//  != add
			if(obj)
				die(ep,MUTEXINITEDEADDRESS,"pthread_mutex_init() returned for address where a mutex already exists (race condition in mutex creation?)\n");
			obj = omh.addObject();
			obj->addr = ep->mtx_addr;
			obj->create_event = ep;
			h.add(ep->mtx_addr,obj);
			break;
		}
		case ev_mutex_create_failed: {
			//object known?
			//  => bad
			if(obj)
				die(ep,MUTEXINITFAILEADDRESS,"pthread_mutex_init failed for address where a mutex already exists\n");
			break;
		}
		case ev_mutex_destroy: {
			//object known ?
			//  => set lifetime
			//  != sat?
			//    => bad
			//    != add create+add+settime
			if(obj)
				obj->destroy_event = ep;
			else {
				if(used_address)
					die(ep,MUTEXDOUBLEDESTROY,"pthread_mutex_destroy() called for mutex which has already been destroyed\n");
				obj = omh.addObject();
				obj->addr = ep->mtx_addr;
				obj->destroy_event = ep;
				h.add(ep->mtx_addr,obj);
			}
			break;
		}
		case ev_mutex_destroyed: {
			//object known?
			//  => forget
			//  != bad
			if(obj)
				h.forget(ep->mtx_addr);
			else
				die(ep,MUTEXDOUBLEDESTROYED,"pthread_mutex_destroy() returned for address where a mutex do not exits (now)\n");
			break;
		}
		case ev_mutex_lock:
		case ev_mutex_trylock: {
			//object known?
			//  != sat?
			//    => bad
			//    != create+add+settime
			if(!obj)
				if(used_address)
					die(ep,MUTEXLOCKDESTROYED,"pthread_mutex_lock/trylock() called for address where a mutex do not exist anymore\n");
				else {
					obj = omh.addObject();
					obj->addr = ep->mtx_addr;
					h.add(ep->mtx_addr,obj);
				}
			break;
		}
		case ev_mutex_locked: {
			//object known?
			//  != bad
			if(!obj) {
				if(used_address)
					die(ep,MUTEXLOCKEDDESTROYED,"pthread_mutex_lock/trylock() returned for address where a mutex do not exist anymore\n");
				else
					die(ep,MUTEXLOCKEDNEXIST,"pthread_mutex_lock/trylock() returned for address where a mutex has never existed (?)\n");
			}
			break;
		}
		case ev_mutex_trylock_failed: {
			//object known?
			//  != bad
			if(!obj)
				die(ep,MUTEXTRYLOCKNEXIST,"pthread_mutex_trylock() failed for address where a mutex do not exist\n");
			break;
		}
		case ev_mutex_unlock: {
			//object known?
			//  != bad
			if(!obj)
				die(ep,MUTEXUNLOCKNEXIST,"pthread_mutex_unlock() called for address where a mutex do not exist, or has never been locked\n");
			break;
		}
		case ev_mutex_unlocked: {
			//object known?
			//  != bad
			if(!obj) {
				if(used_address)
					warn(ep,MUTEXUNLOCKEDDESTROYED,"pthread_mutex_unlock() returned for address where a mutex do not exist anymore (race condition on pthread_mutex_destroy?)\n");
				else
					die(ep,MUTEXUNLOCKEDNEXIST,"pthread_mutex_unlock() returned for address where a mutex has never existed (?)\n");
			}
			break;
		}
		case ev_cond_wait:
		case ev_cond_timedwait: {
			//object known?
			// != sat?
			//   => bad
			//   != register
			if(!obj) {
				if(used_address)
					die(ep,MUTEXCONDWAITDESTROYED,"pthread_cond_wait/timedwait() used a mutex which do not exist (anymore)\n");
				else {
					obj = omh.addObject();
					obj->addr = ep->mtx_addr;
					h.add(ep->mtx_addr,obj);
				}
			}
			break;
		}
		case ev_cond_waited:
		case ev_cond_timedwait_failed: {
			//nothing to do
			break;
		}
		default: 
			die(ep,MUTEX_UNKNOWN_OP,"(internal error) Mutex address given, but unknown event\n");
		} //switch
		eop[i].mtx = obj;
	}
}


static void findConditions(const EventRecord ev[], size_t events,
                           Object **o_cond, size_t *conds,
                           EventObjectPointers eop[]
                          )
{
	AddressObjectHashTable h;
	SeenAddressTable sat;
	ObjectMemoryHandler omh(o_cond,conds,eop,events,h);
	for(size_t i=0; i<events; i++) {
		const EventRecord *ep=ev+i;
		if(ep->cond_addr==0) continue;
		Object *obj = h.find(ep->cond_addr);
		bool used_address = sat.exists(ep->cond_addr);
		sat.add(ep->cond_addr);
		
		switch(ep->event) {
		case ev_cond_name: {
			if(obj) {
				if(obj->name_event)
					warn(ep,CONDALREADYNAMED,
					     "Condition has already been named '%s'\n",
					     obj->name_event->name
					    );
				else
				 	obj->name_event = ep;
			} else {
				obj = omh.addObject();
				obj->addr = ep->cond_addr;
				obj->name_event = ep;
				h.add(ep->cond_addr,obj);
			}
			break;
		}
		case ev_cond_broadcast:
		case ev_cond_signal:
		case ev_cond_wait:
		case ev_cond_timedwait: {
			//object known?
			//  != sat?
			//    => bad
			//    != create+add+settime
			if(!obj)
				if(used_address)
					die(ep,CONDSIGNALDESTROYED,"pthread_cond_broadcast/signal/wait/timedwait() called for address where a condition do not exist anymore");
				else {
					obj = omh.addObject();
					obj->addr = ep->cond_addr;
					h.add(ep->cond_addr,obj);
				}
			break;
		}
		case ev_cond_broadcasted:
		case ev_cond_signalled:
		case ev_cond_waited:
		case ev_cond_timedwait_failed: {
			//object known?
			//  != bad
			if(!obj) {
				if(used_address)
					die(ep,CONDOPEDDESTROYED,"pthread_cond_broadcast/signal/wait/timedwait() returned for address where a condition do not exist anymore");
				else
					die(ep,CONDOPEDNEXIST,"pthread_cond_broadcast/signal/wait/timedwait() returned for address where a condition has never existed (?)");
			}
			break;
		}
		case ev_cond_create: {
			//object known ?
			//  => bad
			if(obj)
				die(ep,CONDINITEADDRESS,"pthread_cond_init() called for address where a condition already exists");
			break;
		}
		case ev_cond_created: {
			//object known?
			//  => bad
			//  != add
			if(obj)
				die(ep,CONDINITEDEADDRESS,"pthread_cond_init() returned for address where a condition already exists (race condition in condition creation?)");
			obj = omh.addObject();
			obj->addr = ep->cond_addr;
			obj->create_event = ep;
			h.add(ep->cond_addr,obj);
			break;
		}
		case ev_cond_create_failed: {
			//object known?
			//  => bad
			if(obj)
				die(ep,CONDINITFAILEADDRESS,"pthread_cond_init failed for address where a condition already exists");
			break;
		}
		case ev_cond_destroy: {
			//object known ?
			//  => set lifetime
			//  != sat?
			//    => bad
			//    != add create+add+settime
			if(obj)
				obj->destroy_event = ep;
			else {
				if(used_address)
					die(ep,CONDDOUBLEDESTROY,"pthread_cond_destroy() called for condition which has already been destroyed");
				obj = omh.addObject();
				obj->addr = ep->cond_addr;
				obj->destroy_event = ep;
				h.add(ep->cond_addr,obj);
			}
			break;
		}
		case ev_cond_destroyed: {
			//object known?
			//  => forget
			//  != bad
			if(obj)
				h.forget(ep->cond_addr);
			else
				die(ep,CONDDOUBLEDESTROYED,"pthread_cond_destroy() returned for address where a condition do not exist (now)");
			break;
		}
		default:
			die(ep,COND_UNKNOWN_OP,"(internal error) condition address set, but unknown event type\n");
		} //switch
		eop[i].cond = obj;
	}
}


static void findThreads(const EventRecord ev[], size_t events,
                        Object **o_thr, size_t *thrs,
                        EventObjectPointers eop[]
                       )
{
	AddressObjectHashTable h; //slight misuse of the term "address"
	ObjectMemoryHandler omh(o_thr,thrs,eop,events,h);
	bool first_thread=true;
	for(size_t i=0; i<events; i++) {
		const EventRecord *ep=ev+i;
		Object *obj = h.find(ep->tid);
		if(!obj) {
			obj = omh.addObject();
			obj->addr = ep->tid;
			if(!first_thread)
				obj->create_event = ep;
			h.add(ep->tid,obj);
			first_thread = false;
		}
		switch(ep->event) {
		case ev_thread_created: {
			//info is tid
			Object *created_thread = h.find(ep->info);
			if(!created_thread) {
				created_thread = omh.addObject();
				created_thread->addr = ep->info;
				created_thread->create_event = ep;
				h.add(ep->info,created_thread);
			}
			break;
		}
		case ev_thread_exit: {
			obj->destroy_event = ep;
			h.forget(ep->tid);
			break;
		}
		case ev_thread_joined: {
			if(ep->info==0) {
				//scan backward after ev_thread_join to find tid
				for(size_t j=i; j--!=0; ) {
					if(ev[j].tid==ep->tid &&
					   ev[j].event == ev_thread_join)
					{
						Object *joined_thread = h.find(ev[j].info);
						if(joined_thread) {
							joined_thread->destroy_event = ep;
							h.forget(joined_thread->addr);
						} else {
							//warn(ep,"Thread joined a unknown thread (tid=%08lx)\n",ev[j].info);
						}
						break;
					}
				}
			}
			break;
		}
		case ev_thread_name: {
			if(obj->name_event)
				warn(ep,THREADALREADYNAMED,
				     "Thread has already been named '%s'\n",
				     obj->name_event->name
				    );
			else
			 	obj->name_event = ep;
			break;
		}
		default:
			;
		} //switch
		eop[i].thr = obj;
	}
}


bool preAnalyze(const EventRecord ev[], size_t events,
                Object **o_mtx, size_t *mtxs,
                Object **o_cond, size_t *conds,
                Object **o_thr, size_t *thrs,
                EventObjectPointers **eop
               )
{
	OutputChapterStart("pre-analysis");
	
	*eop = (EventObjectPointers*)malloc(sizeof(EventObjectPointers)*events);
	memset(*eop,0,sizeof(EventObjectPointers)*events);
	
	*o_mtx = 0;
	*mtxs = 0;
	
	*o_cond = 0;
	*conds = 0;
	
	OutputSectionStart("mutex-discovery");
	findMutexes(ev,events,o_mtx,mtxs,*eop);
	OutputParagraphStart();
	fprintf(fp_output,"Found %ld mutexe%s\n",(long)*mtxs,*mtxs==1?"":"s");
	OutputParagraphEnd();
	OutputSectionEnd("mutex-discovery");
	
	OutputSectionStart("condition-discovery");
	findConditions(ev,events,o_cond,conds,*eop);
	OutputParagraphStart();
	fprintf(fp_output,"Found %ld condition%s\n",(long)*conds,*conds==1?"":"s");
	OutputParagraphEnd();
	OutputSectionEnd("condition-discovery");
	
	OutputSectionStart("thread-discovery");
	findThreads(ev,events,o_thr,thrs,*eop);
	OutputParagraphStart();
	fprintf(fp_output,"Found %ld thread%s\n",(long)*thrs,*thrs==1?"":"s");
	OutputParagraphEnd();
	OutputSectionEnd("thread-discovery");
	
	OutputChapterEnd("pre-analysis");
	return true;
}
